/***********************************************************************************************************************
 *
 * blueBill Mobile - Android - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it/mobile
 * SCM: https://java.net/hg/bluebill-mobile~android-src
 *
 **********************************************************************************************************************/
package it.tidalwave.mobile.android.ui;

import javax.annotation.Nonnull;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import org.openide.util.Lookup;
import android.app.Activity;
import android.content.ComponentName;
import android.content.Context;
import android.content.Intent;
import it.tidalwave.util.ui.FlowController;
import it.tidalwave.util.Parameters;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import java.util.ArrayList;
import java.util.List;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
public abstract class AndroidFlowController implements FlowController
  {
    private static final Logger log = LoggerFactory.getLogger(AndroidFlowController.class);

    /** Flags nextStep() to use the standard Intent mechanism for finding the next Activity, instead of relying of
     *  the registered sequences. It's still important to use nextStep() instead of Activity.startActivity() in
     *  order to propagate the FlowController class name and requestCode. */
    public static final Object USE_INTENT_FILTER = "USE_INTENT_FILTER";

    private static int nextRequestCode = 1000;

    /* package */ static final String CONTROL_FLOW_CLASS_NAME = AndroidFlowController.class.getName() + ".controlFlowClassName";

    private static final Map<Integer, Class<? extends AndroidFlowController>> REQUEST_CODE_TO_CONTROL_FLOW_MAP = new HashMap<Integer, Class<? extends AndroidFlowController>>();
    private static final Map<Class<? extends AndroidFlowController>, Integer> CONTROL_FLOW_MAP_TO_REQUEST_CODE = new HashMap<Class<? extends AndroidFlowController>, Integer>();

    private final Map<Class<? extends Activity>, List<ConditionAndActivityPair>> transitionMap =
            new HashMap<Class<? extends Activity>, List<ConditionAndActivityPair>>();

    private Class<? extends Activity> initialActivity;

    protected Activity activity;

    private final int requestCode;

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private static class ConditionAndActivityPair
      {
        @Nonnull
        protected final Condition condition;

        @Nonnull
        protected final Class<? extends Activity> activityClass;

        public ConditionAndActivityPair (final @Nonnull Condition condition,
                                         final @Nonnull Class<? extends Activity> activityClass)
          {
            this.condition = condition;
            this.activityClass = activityClass;
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public class FlowBuilder
      {
        private final Class<? extends Activity> from;
        private Condition condition;

        protected FlowBuilder (final @Nonnull Class<? extends Activity> from)
          {
            this.from = from;
          }

        @Nonnull
        public FlowBuilder completes()
          {
            return completesWith(TRUE);
          }

        @Nonnull
        public synchronized FlowBuilder completesWith (final @Nonnull Condition condition)
          {
            this.condition = condition;
            return this;
          }

        @Nonnull
        public FlowBuilder completesWithout (final @Nonnull Condition condition)
          {
            return completesWith(not(condition));
          }

        public void thenForwardTo (final @Nonnull Class<? extends Activity> to)
          {
            List<ConditionAndActivityPair> pairs = transitionMap.get(from);

            if (pairs == null)
              {
                pairs = new ArrayList<ConditionAndActivityPair>();
                transitionMap.put(from, pairs);
              }

            pairs.add(new ConditionAndActivityPair(condition, to));
          }

        private final Condition TRUE = new Condition()
          {
            public boolean compute (final @Nonnull Object... args)
              {
                return true;
              }
         };

        @Nonnull
        private Condition not (final @Nonnull Condition condition)
          {
            return new Condition()
              {
                public boolean compute (final @Nonnull Object... args)
                  {
                    return !condition.compute(args);
                  }
              };
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    // TODO: I'd like to have the requestCode auto-initialized, but I'm not sure it always works. What about the
    // activity gets unloaded, thus we lose the CONTROL_FLOW_MAP_TO_REQUEST_CODE? When the FlowController is used again
    // it would recreate a requestCode, but we don't have a guarantee that FlowController subclasses are reinstantiated
    // in the same order, so they'd get different requestCodes. Must understand better when classes are unloaded to
    // understand whether it can happen or not.

    protected AndroidFlowController()
      {
        synchronized (CONTROL_FLOW_MAP_TO_REQUEST_CODE)
          {
            final Class<? extends AndroidFlowController> clazz = getClass();

            if (CONTROL_FLOW_MAP_TO_REQUEST_CODE.containsKey(clazz))
              {
                requestCode = CONTROL_FLOW_MAP_TO_REQUEST_CODE.get(clazz);
              }
            else
              {
                requestCode = nextRequestCode++;
                REQUEST_CODE_TO_CONTROL_FLOW_MAP.put(requestCode, clazz);
                CONTROL_FLOW_MAP_TO_REQUEST_CODE.put(clazz, requestCode);
                log.info("{} registered to requestCode: {}", clazz, requestCode);
              }
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected void startFrom (final @Nonnull Class<? extends Activity> initialActivity)
      {
        this.initialActivity = initialActivity;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    protected FlowBuilder when (final @Nonnull Class<? extends Activity> initialActivity)
      {
        return new FlowBuilder(initialActivity);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public int getRequestCode()
      {
        return requestCode;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public Class<? extends Activity> getInitialActivity()
      {
        return initialActivity;
      }

    /*******************************************************************************************************************
     *
     * Starts a new {@link FlowController}.
     *
     * @param  controlFlowClass  the {@link FlowController} to start
     *
     ******************************************************************************************************************/
    public final void start (final @Nonnull Class<? extends AndroidFlowController> controlFlowClass)
      {
        try
          {
            log.info("startActivity({}) - {}", controlFlowClass, activity);
            final AndroidFlowController controlFlow = controlFlowClass.newInstance();
            final Class<? extends Activity> initialActivity = controlFlow.getInitialActivity();
            final int requestCode = controlFlow.getRequestCode();
            log.info(">>>> initialActivity: {}, requestCode: {}", initialActivity, requestCode);
            final Intent intent = new Intent(getContext(), initialActivity);
            intent.putExtra(CONTROL_FLOW_CLASS_NAME, controlFlowClass.getName());
            activity.startActivityForResult(intent, requestCode);
          }
        catch (Exception e)
          {
            throw new RuntimeException(e);
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public static AndroidFlowController forActivity (final @Nonnull Activity activity)
      {
        return new AndroidFlowControllerLazyProxy(activity);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    /* package */ static AndroidFlowController forActivity2 (final @Nonnull Activity activity)
      {
        Parameters.checkNonNull(activity, "activity");

        try
          {
            final String className = activity.getIntent().getStringExtra(CONTROL_FLOW_CLASS_NAME);
            assert className != null : "Sorry, I've lost the CONTROL_FLOW_CLASS_NAME!";
            final AndroidFlowController controlFlow = (AndroidFlowController)Class.forName(className).newInstance();
            controlFlow.activity = activity;
            return controlFlow;
          }
        catch (Exception e)
          {
            throw new RuntimeException(e);
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public static int findRequestCode (final @Nonnull Class<? extends AndroidFlowController> controlFlowClass)
      {
        if (!CONTROL_FLOW_MAP_TO_REQUEST_CODE.containsKey(controlFlowClass))
          {
            try // try to register it
              {
                controlFlowClass.newInstance();
              }
            catch (InstantiationException e)
              {
                throw new RuntimeException(e);
              }
            catch (IllegalAccessException e)
              {
                throw new RuntimeException(e);
              }
          }

        if (!CONTROL_FLOW_MAP_TO_REQUEST_CODE.containsKey(controlFlowClass))
          {
            throw new RuntimeException("ControlFlow not registered: " + controlFlowClass);
          }

        return CONTROL_FLOW_MAP_TO_REQUEST_CODE.get(controlFlowClass);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public void finish()
      {
        throw new UnsupportedOperationException("To be implemented");
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public void toNextStep (final @Nonnull Object ... args)
      {
        toNextStep(new Intent(), args);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public void toNextStep (final @Nonnull Intent intent, final @Nonnull Object ... args)
      {
        log.info("toNextStep({}, {})", intent, Arrays.toString(args));

        if (activity == null)
          {
            throw new IllegalStateException("ControlFlow must have been retrevied by ControlFlow.for_()");
          }

        if (Arrays.asList(args).contains(USE_INTENT_FILTER))
          {
            log.info(">>>> using the intent for starting the next Activity...");
            final Intent intentDecorator = new Intent(intent);
            intentDecorator.putExtra(CONTROL_FLOW_CLASS_NAME, getClass().getName());
            activity.startActivityForResult(intentDecorator, requestCode);
            return;
          }

        Class<? extends Activity> nextActivityClass = null;
        final ComponentName componentName = intent.getComponent();

        if (componentName == null)
          {
            final List<ConditionAndActivityPair> pairs = transitionMap.get(activity.getClass());

            if (pairs != null)
              {
                for (final ConditionAndActivityPair pair : pairs)
                  {
                    if (pair.condition.compute(args))
                      {
                        nextActivityClass = pair.activityClass;
                        break;
                      }
                  }
              }
          }
        else
          {
            final String explicitNextActivityClassName = componentName.getClassName();
            log.info(">>>> explicit next activity: {}", explicitNextActivityClassName);

            try
              {
                nextActivityClass = (Class<? extends Activity>)Class.forName(explicitNextActivityClassName);
              }
            catch (ClassNotFoundException e2)
              {
                throw new RuntimeException(e2);
              }
          }

        if (nextActivityClass == null)
          {
            log.info(">>>> it was the final step, calling finish()...");
            activity.setResult(Activity.RESULT_OK);
            activity.finish();
          }
        else
          {
            log.info(">>>> next Activity is: {}", nextActivityClass);
            final Intent intentDecorator = new Intent(intent);
            intentDecorator.setClass(getContext(), nextActivityClass);
            intentDecorator.putExtra(CONTROL_FLOW_CLASS_NAME, getClass().getName());
            activity.startActivityForResult(intentDecorator, requestCode);
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private static Context getContext()
      {
        return Lookup.getDefault().lookup(Context.class);
      }
  }
