/***********************************************************************************************************************
 *
 * blueBill Mobile - Android - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it/mobile
 * SCM: https://java.net/hg/bluebill-mobile~android-src
 *
 **********************************************************************************************************************/
package it.tidalwave.ui.android.view;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import it.tidalwave.util.thread.annotation.ThreadConfined;
import android.content.Context;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;
import static it.tidalwave.util.thread.ThreadType.*;

/***********************************************************************************************************************
 *
 * Facility which provides a convenient implementation of a idiomatic procedure of Android for rendering elements in 
 * lists via {@link android.widget.Adapter}.
 * 
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
public class AdapterViewHelper
  {
    public static interface Updater<T extends View>
      {
        public void update (@Nonnull T view);
      }
    
    @Nonnull
    private final LayoutInflater layoutInflater;

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public AdapterViewHelper (final @Nonnull Context context)
      { 
        this.layoutInflater = LayoutInflater.from(context);
      }
    
    /*******************************************************************************************************************
     *
     * Prepares a {@link View} to be used with overridden methods
     * {@link android.widget.Adapter#getView(int, android.view.View, android.view.ViewGroup)}
     * 
     * @param  layoutId         the id of the layout to be used
     * @param  componentViewId  the id of the component {@code View} that must be updated
     * @param  view             the view as provided by the {@code Adapter.getView()} method
     * @param  updater          the callback that will be used for updating the component {@code View}
     * @return                  the {@code View} that must be returned in the {@code Adapter.getView()} method
     *
     ******************************************************************************************************************/
    @Nonnull @ThreadConfined(type=UI)
    public <T extends View> View prepareView (final int layoutId, 
                                              final int componentViewId, 
                                              @CheckForNull View view, 
                                              final @Nonnull ViewGroup viewGroup, 
                                              final @Nonnull Updater<T> updater) 
      {
        T component;

        if (view == null)
          {
            view = layoutInflater.inflate(layoutId, null);
            component = (T)view.findViewById(componentViewId);
            view.setTag(component);
          }
        else
          {
            component = (T)view.getTag();
          }
        
        updater.update(component);
        
        return view;
      }   
  }
